import ctypes, signal, threading, multiprocessing
from collections import deque

class KillException(BaseException):
	pass

class Thread(threading.Thread):
	'''A thread class that supports raising exception in the thread from
	   another thread.
	'''
	def get_id(self):
		if not self.isAlive(): return #If thread is dead
		if hasattr(self, "tid"): return self.tid #earlier grab
		# check threading module _active dict
		for tid, obj in threading._active.items():
			if obj is self:
				self.tid = tid
				return tid
		raise AssertionError("No thread ID")
		
	def stop(self):
		tid = self.get_id()
		if tid:
			res = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid),
															 ctypes.py_object(KillException))
			if res == 0:
				raise ValueError("invalid thread id")
			elif res != 1:
				# "if it returns a number greater than one, you're in trouble,
				# and you should call it again with exc=NULL to revert the effect"
				ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid), None)
				raise SystemError("PyThreadState_SetAsyncExc failed")


class threader_base:
	def __init__(self):
		self.queue = deque()
		self.threads = deque()

	def add(self, callback, arglist: list = []):
		t = Thread(target = callback, args = arglist)
		self.queue.append(t)

	def start(self):		
		self._clear_dead_threads()
		for t in list(self.queue):
			self.queue.remove(t)
			started = False
			while not started:
				try:
					t.start()
					started = True
				except RuntimeError:
					pass

			self.threads.append(t)

	def join(self):
		for t in self.threads:
			t.join()

	#Not to be called by user
	def _clear_dead_threads(self):
		if self.threads:
			for t in list(self.threads):
				if not t.is_alive():
					self.threads.remove(t)

	def kill(self):
		threader_base.kill(self)
		print("Killing all worker threads.")
		for t in self.threads:
			print(f"Stopping {t.name}")
			t.stop()

class Session(threader_base):
	def __init__(self):
		threader_base.__init__(self)

class Threader(threader_base):
	def __init__(self):
		self.queue = deque()
		self.threads = deque()
		self.sessions = []

		signal.signal(signal.SIGINT, self.exit)
		signal.signal(signal.SIGTERM, self.exit)

	def kill(self):
		threader_base.kill(self)
		for s in self.sessions:
			s.kill()

	def add_session(self):
		s = Session()
		self.sessions.append(s)
		return s

	def end_session(self, session):
		self.sessions.remove(session)

	def exit(self, event=None):
		self.kill()